/*
================================================================================
 File: boot.c
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       boot
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#include <stddef.h>

#include "boot.h"
#include "boot_config.h"
#include "fat32.h"
#include "sd_spi.h"
#include "flash.h"
#include "user_app_test.h"

#include "stm32f10x_map.h"
#include "stm32f10x_gpio.h"
#include "stm32f10x_rcc.h"

/* --------- PWR registers bit address in the alias region ---------- */
#define PWR_OFFSET               (PWR_BASE - PERIPH_BASE)

/* --- CR Register ---*/
/* Alias word address of DBP bit */
#define CR_OFFSET                (PWR_OFFSET + 0x00)
#define DBP_BitNumber            0x08
#define CR_DBP_BB                (PERIPH_BB_BASE + (CR_OFFSET * 32) + (DBP_BitNumber * 4))

#define USE_LED				1



#define SD_SPI_MOSI_PORT				GPIOB
#define SD_SPI_MOSI_PIN					GPIO_Pin_15

#define SD_SPI_MISO_PORT				GPIOB
#define SD_SPI_MISO_PIN					GPIO_Pin_14

#define SD_SPI_SCK_PORT					GPIOB
#define SD_SPI_SCK_PIN					GPIO_Pin_13

#define SD_SPI_NSS_PORT					GPIOB
#define SD_SPI_NSS_PIN					GPIO_Pin_12


#define LED1_PORT						GPIOC
#define LED1_PIN						GPIO_Pin_12
#define LED1_PIN_MODE					GPIO_Mode_Out_PP

#define WakeUp_PORT			GPIOA
#define WakeUp_PIN			GPIO_Pin_0
#define WakeUp_PIN_MODE		GPIO_Mode_IPD

#define LED_OFF()		LED1_PORT->BSRR = LED1_PIN
#define LED_ON()		LED1_PORT->BRR = LED1_PIN

static uint16_t boot_page_size;
static uint8_t boot_read_buffer[260];
//static FAT_UserDesc FC;
static SD_SPI_MSD SD_SPI;


#if(USE_LED)
void led_init(void) {
	GPIO_InitTypeDef gpio_led;
	
	gpio_led.GPIO_Pin = LED1_PIN;
	gpio_led.GPIO_Speed = GPIO_Speed_50MHz;
	gpio_led.GPIO_Mode = LED1_PIN_MODE;
	GPIO_Init(LED1_PORT, &gpio_led);
}
#endif


void spi_init(void) {
	GPIO_InitTypeDef gpio;

	gpio.GPIO_Pin = SD_SPI_MOSI_PIN;
	gpio.GPIO_Speed = GPIO_Speed_50MHz;
	gpio.GPIO_Mode = GPIO_Mode_AF_PP;
	GPIO_Init(SD_SPI_MOSI_PORT, &gpio);

	gpio.GPIO_Pin = SD_SPI_MISO_PIN;
	GPIO_Init(SD_SPI_MISO_PORT, &gpio);

	gpio.GPIO_Pin = SD_SPI_SCK_PIN;
	GPIO_Init(SD_SPI_SCK_PORT, &gpio);
}


static FILE file_desc;
void boot_process(uint8_t *dev_name) {
	uint32_t result;
	uint32_t address_shift = 0;
	uint16_t page_piece = 0;

	if(!sfopen(&file_desc, (const char*)dev_name, "r")) {
		return;
	}
	if(!sfopen(&file_desc, "stm32f10x/file_name.txt", "r")) {
		return;
	}
	
	fread(boot_read_buffer, 1, 260, &file_desc);

	fclose(&file_desc);
	
	if(!sfopen(&file_desc, (const char *)boot_read_buffer, "r")) {
		return;
	}
	
#if(USE_LED)
	LED_ON();
#endif

	FlashUnlock();
	do {
		if(!page_piece) {
			if (FlashErasePage(DEF_APP_ADDRESS + address_shift))
			{
				break;
			}
		}

		result = fread(boot_read_buffer, 1, 256, &file_desc);

		if (FlashWritePage(DEF_APP_ADDRESS + address_shift, boot_read_buffer, result))
		{
			break;
		}

		address_shift += result;
		page_piece++;

		if((page_piece << 8) == boot_page_size) {
			page_piece = 0;
		}
	}while(result == 256);
	FlashLock();

#if(USE_LED)
	LED_OFF();
#endif

	fclose(&file_desc);
}


#define __RCC_APB1Periph_ALL               ((uint32_t)0x3AFEC83F)
#define __RCC_APB2Periph_ALL               ((uint32_t)0x0000FFFD)
/* Alias word address of PLLON bit */
#define PLLON_BitNumber           0x18
#define CR_PLLON_BB               (PERIPH_BB_BASE + (CR_OFFSET * 32) + (PLLON_BitNumber * 4))

void boot_rcc_deinit_all(void) {
	// Full peripherial reset
	RCC->APB1RSTR |= __RCC_APB1Periph_ALL;
	RCC->APB2RSTR |= __RCC_APB2Periph_ALL;
	// Disable APB2 Peripheral Reset
	RCC->APB2RSTR = 0x00000000;

	// Disable APB1 Peripheral Reset
	RCC->APB1RSTR = 0x00000000;

	// FLITF and SRAM Clock ON
	RCC->AHBENR = 0x00000014;

	// Disable APB2 Peripheral Clock
	RCC->APB2ENR = 0x00000000;

	// Disable APB1 Peripheral Clock
	RCC->APB1ENR = 0x00000000;
    
	// Set HSI as system clock
    RCC->CFGR &= ~((uint32_t)0x3);

    // Disable PLL
    *(vu32 *) CR_PLLON_BB = (u32)0;

    // Reset all PLL and other parameters
    RCC->CFGR = 0;
    
    // Reset all RCC parameters
    RCC->CR = 0x83;
    
    // Disable all interrupts
    RCC->CIR = 0x00000000;
}


void boot_force(void) {
#if(USE_LED)
	led_init();
	LED_OFF();
#endif
	
	if(*FLASH_SIZE_REG >= 256) {
		boot_page_size = 2048;
	}
	else {
		boot_page_size = 1024;
	}
	
    FAT_Init();
    
//    FC.FAT_ConnectEventHandler = FAT_ConnectEvent;
//    FC.FAT_DisconnectEventHandler = 0;
//    FAT_UserInstall(&FC);

    spi_init();

    SD_SPI_init(&SD_SPI, SPI2, GPIOB, GPIO_Pin_12);
    
    if(MSM_GetNumInstalledDevices()) {
    	boot_process(MSM_GetDeviceByID(0)->string_descriptor);
    }
	
	boot_rcc_deinit_all();
}

void boot_try(void) {
	GPIO_InitTypeDef gpio;

	//  WakeUp pin initialize
	gpio.GPIO_Pin = WakeUp_PIN;
	gpio.GPIO_Speed = GPIO_Speed_50MHz;
	gpio.GPIO_Mode = WakeUp_PIN_MODE;
	GPIO_Init(WakeUp_PORT, &gpio);

#if(BOOT_CALL_BY_RESET_SUPPORT)

	RCC->APB1ENR |= RCC_APB1Periph_BKP | RCC_APB1Periph_PWR;
	RCC->APB1RSTR &= ~((uint32_t)(RCC_APB1Periph_BKP | RCC_APB1Periph_PWR));
	// Enable write access to Backup registers
	*(vu32 *) CR_DBP_BB = (u32)1;

	if((WakeUp_PORT->IDR & WakeUp_PIN) || !is_app_correct() || (BKP->DR1 == 0x159D)) {
		if(BKP->DR1 == 0x159D) {
			BKP->DR1 = 0;
		}
		boot_force();
	}
	
#else

	if((WakeUp_PORT->IDR & WakeUp_PIN) || !is_app_correct()) {
		boot_force();
	}

#endif
	
	boot_rcc_deinit_all();
}
