/*
================================================================================
 File: sd_spi.c
--------------------------------------------------------------------------------
 Author:        Piotr Wojtowicz
 Purpose:       SD library uses SPI interface MSM compatible
 Compiler:      ANSI C
 Hardware:      
 Doc:
--------------------------------------------------------------------------------
 Includes:      
--------------------------------------------------------------------------------
 Comments:                      UWAGA!!!
                Oprogramowanie to udostępnione jest do celów poglądowych,
                nie jest przetestowane pod kątem bezpieczeństwa ani spełnienia
                jakichkolwiek norm wymaganych aby urządzenie używające go
                mogło być dopuszczone do użytku, dlatego używasz go
                na własną odpowiedzialność a jego autor nie ponosi żadnego
                typu odpowiedzialności za jakiekolwiek skutki jego użycia.
                Jeżeli nie zgadzasz się z powyższym warunkiem proszę
                o nieużywanie tego oprogramowania.
================================================================================
*/

#include "sd_spi.h"


#include "crc7.h"
#include "crc16.h"



typedef struct {
	// Manufacturer ID
	uint8_t MID;

	// OEM Application ID
	uint16_t OID;

	//Product name
	uint8_t PNM[5];

	// Product revision
	uint8_t PRV;

	//Product serial number
	uint32_t PSN;

	// Manufacturing date
	uint16_t MDT;
}SD_CID_register;



static uint8_t SD_SPI_Initialize(SD_SPI_MSD *SD);
static void SD_SPI_SendCommand(SPI_TypeDef *SD_SPI, uint8_t cmd_index, uint32_t argument);
static void SD_SPI_Get1ByteToken(SPI_TypeDef *SD_SPI, uint8_t *response);
static void SD_SPI_GetR3R7(SPI_TypeDef *SD_SPI, uint8_t *response);
static int SD_SPI_GetCID(SD_SPI_MSD *SD, SD_CID_register *CID);
static int SD_SPI_GetCSD(SD_SPI_MSD *SD, uint8_t *CSD);
#if(MSM_READ_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_ReadBlock(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_WRITE_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_WriteBlock(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_VERIFY_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_VerifyBlock(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif

#if(MSM_ERASE_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_EraseBlock(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count);
#endif


#define CR1_SPE_Set          ((uint16_t)0x0040)


// sterowanie linia CE
// ENABLE
#define SD_SPI_CE_ENABLE(SD)		(SD)->GPIO_NSS->BRR = (SD)->GPIO_Pin_NSS

// DISABLE
#if(SD_MINIMUM_SIZE_SUPPORT != 1)
#define SD_SPI_CE_DISABLE(SD)	(SD)->GPIO_NSS->BSRR = (SD)->GPIO_Pin_NSS;\
						while(!((SD)->SD_SPI->SR & SPI_FLAG_TXE));\
						(SD)->SD_SPI->DR = 0xFF;
#else
static void SD_SPI_CE_DISABLE(SD_SPI_MSD *SD) {
	SD->GPIO_NSS->BSRR = SD->GPIO_Pin_NSS;
	while(!(SD->SD_SPI->SR & SPI_FLAG_TXE));
	SD->SD_SPI->DR = 0xFF;
}
#endif





void SD_SPI_init(SD_SPI_MSD *SD, SPI_TypeDef *SD_SPI, GPIO_TypeDef* GPIO_NSS, uint16_t GPIO_Pin_NSS) {
	GPIO_InitTypeDef gpio;
	SPI_InitTypeDef SPI_InitStructure;
	SD_CID_register CID;
	uint8_t CSD[15];
	uint8_t READ_BL_LEN;
	uint8_t C_SIZE_MULT;
	uint32_t counter;
	
	union {
		uint32_t uint32;
		uint8_t uint8[4];
	}C_SIZE;
	
	SD->SD_SPI = SD_SPI;
	SD->GPIO_NSS = GPIO_NSS;
	SD->GPIO_Pin_NSS = GPIO_Pin_NSS;
	
	gpio.GPIO_Pin = GPIO_Pin_NSS;
	gpio.GPIO_Speed = GPIO_Speed_50MHz;
	gpio.GPIO_Mode = GPIO_Mode_Out_PP;
	GPIO_Init(GPIO_NSS, &gpio);

	// set NSS pin High
	GPIO_NSS->BSRR = GPIO_Pin_NSS;

	// enable peripheral clock and reset
	// SPI1
	if(((uint32_t)SD_SPI) == ((uint32_t)SPI1)) {
		RCC->APB2ENR |= 1 << 12;
		RCC->APB2RSTR |= 1 << 12;
		RCC->APB2RSTR &= ~((uint32_t)(1 << 12));
	}
	// SPI2
	if(((uint32_t)SD_SPI) == ((uint32_t)SPI2)) {
		RCC->APB1ENR |= 1 << 14;
		RCC->APB1RSTR |= 1 << 14;
		RCC->APB1RSTR &= ~((uint32_t)(1 << 14));
	}


	SPI_InitStructure.SPI_Direction = SPI_Direction_2Lines_FullDuplex;
	SPI_InitStructure.SPI_Mode = SPI_Mode_Master;
	SPI_InitStructure.SPI_DataSize = SPI_DataSize_8b;
	SPI_InitStructure.SPI_CPOL = SPI_CPOL_High;
	SPI_InitStructure.SPI_CPHA = SPI_CPHA_2Edge;
	SPI_InitStructure.SPI_NSS = SPI_NSS_Soft;
	SPI_InitStructure.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_256;
	SPI_InitStructure.SPI_FirstBit = SPI_FirstBit_MSB;
	SPI_InitStructure.SPI_CRCPolynomial = 7;

	// initialize peripheral
	SPI_Init(SD_SPI, &SPI_InitStructure);
	// enable peripheral
	SPI_Cmd(SD_SPI, ENABLE);

	// odczekanie 80 cykli zegara SPI
	for(counter = 0; counter < 10; counter++) {
		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = 0xFF;
	}

	if(!(SD_SPI_Initialize(SD))) {
		if(!(SD_SPI_GetCID(SD, &CID))) {
			if(!(SD_SPI_GetCSD(SD, CSD))) {

				SPI_Cmd(SD_SPI, DISABLE);
#if((CRC7_LIB_MAX_SPEED_SUPPORT == 1) && (CRC16_LIB_MAX_SPEED_SUPPORT == 1))
				SPI_InitStructure.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_16;
#else
				SPI_InitStructure.SPI_BaudRatePrescaler = SPI_BaudRatePrescaler_4;
#endif
				// initialize peripheral
				SPI_Init(SD_SPI, &SPI_InitStructure);
				// enable peripheral
				SPI_Cmd(SD_SPI, ENABLE);//*/
				
				
				// indeksy sa mniejsze o 1 od wynikajacych z dokumentacji
				// poniewaz tablica CSD nie zawiera pierwszzego bajtu - CRC
				// versoin 1.0
				if(!(CSD[14] & 0xC0)) {
					C_SIZE.uint8[0] = CSD[6];
					C_SIZE.uint8[1] = CSD[7];
					C_SIZE.uint8[2] = CSD[8] & 0x03;
					C_SIZE.uint8[3] = 0;
					C_SIZE.uint32 >>= 6;

					
					C_SIZE_MULT = CSD[4] >> 7;
					C_SIZE_MULT |= (CSD[5] << 1) & 0x06;

					READ_BL_LEN = CSD[9] & 0xF;

					C_SIZE.uint32++;
					SD->SD_Desc.area_size_L = C_SIZE.uint32 << (2 + C_SIZE_MULT + READ_BL_LEN);
					SD->SD_Desc.area_size_H = C_SIZE.uint32 >> (30 - (C_SIZE_MULT + READ_BL_LEN));

					// here we ignore SDCardInfo.SD_csd.RdBlockLen field, because it gives
					// maximum read/write block size. This value can be just one of three
					// following values: 0x9, 0xA or 0xB, that indicates the maximum read/write
					// block size can be 512, 1024 or 2048 bytes for Standard SD card
					// (for HCSD it is always 512). Minimum is always 1-byte (for Standard SD)
					// so we can set 512 to be compatible with many file system libraries
#if(MSM_READ_SUPPORT == 1)
					SD->SD_Desc.read_sector_size = 512;//1 << READ_BL_LEN;
#endif

#if(MSM_VERIFY_SUPPORT == 1)
					SD->SD_Desc.verify_sector_size = 512;//1 << READ_BL_LEN;
#endif

#if(MSM_WRITE_SUPPORT == 1)
					SD->SD_Desc.write_sector_size = 512;//1 << READ_BL_LEN;
#endif

#if(MSM_ERASE_SUPPORT == 1)
					if(CSD[4] & 0x40) {
						SD->SD_Desc.erase_sector_size = 512;
					}
					else {
						SD->SD_Desc.erase_sector_size = (CSD[3] >> 7) & 0x1;
						SD->SD_Desc.erase_sector_size |= (CSD[4] << 1) & 0x7E;
						SD->SD_Desc.erase_sector_size++;
						SD->SD_Desc.erase_sector_size <<= READ_BL_LEN;
					}
#endif
					
					SD->SD_Desc.string_descriptor = SD->SD_Descriptor;

					SD->SD_Descriptor[0] = '/';
					SD->SD_Descriptor[1] = 'S';
					SD->SD_Descriptor[2] = 'D';
					SD->SD_Descriptor[3] = '_';
					SD->SD_Descriptor[4] = 'S';
					SD->SD_Descriptor[5] = 'P';
					SD->SD_Descriptor[6] = 'I';
					if(((uint32_t)SD_SPI) == ((uint32_t)SPI1)) {
						SD->SD_Desc.string_descriptor[7] = '1';
					}
					if(((uint32_t)SD_SPI) == ((uint32_t)SPI2)) {
						SD->SD_Desc.string_descriptor[7] = '2';
					}
					SD->SD_Descriptor[8] = 0;
					
/*					report((u8*)"area_size_L", SD_Desc.area_size_L);
					report((u8*)"area_size_H", SD_Desc.area_size_H);

					report((u8*)"read_sector_size", SD_Desc.read_sector_size);
					report((u8*)"write_sector_size", SD_Desc.write_sector_size);
					report((u8*)"verify_sector_size", SD_Desc.verify_sector_size);
					report((u8*)"erase_sector_size", SD_Desc.erase_sector_size);//*/


#if(MSM_READ_SUPPORT == 1)
					SD->SD_Parameters.ReadBlockHandler = SD_SPI_ReadBlock;
#endif

#if(MSM_WRITE_SUPPORT == 1)
					SD->SD_Parameters.WriteBlockHandler = SD_SPI_WriteBlock;
#endif

#if(MSM_VERIFY_SUPPORT == 1)
					SD->SD_Parameters.VerifyBlockHandler = SD_SPI_VerifyBlock;
#endif

#if(MSM_ERASE_SUPPORT == 1)
					SD->SD_Parameters.EraseBlockHandler = SD_SPI_EraseBlock;
#endif


					MSMD_Install(&(SD->SD_Desc), &(SD->SD_Parameters));
				}
			}
		}
	}
}




#if(MSM_READ_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_ReadSector(SD_SPI_MSD *SD, uint8_t *data, uint32_t addr_L, uint32_t addr_H) {
	uint8_t response = 0xFF;
	uint16_t crc = 0;
	uint32_t data_left = SD->SD_Desc.read_sector_size;
	SPI_TypeDef *SD_SPI = SD->SD_SPI;
	
	SD_SPI_CE_ENABLE(SD);
	SD_SPI_SendCommand(SD_SPI, 17, addr_L);
	SD_SPI_Get1ByteToken(SD_SPI, &response);
	// SD card is ready and there was no error in the command
	if(!response) {
		SD_SPI_Get1ByteToken(SD_SPI, &response);
		// Single block read - at fist we receive 1-byte data token
		// that should be 0xFE - start block
		if(response == 0xFE) {
			while(data_left){
				if(SD_SPI->SR & SPI_FLAG_TXE) {
					SD_SPI->DR = 0xFF;
				}
				if(SD_SPI->SR & SPI_FLAG_RXNE) {
					data_left--;
					*data = (uint8_t)(SD_SPI->DR);
					crc = crc16(crc, *data);
					data++;
				}
			}

			// get CRC
			response = 2;
			do {
				if(SD_SPI->SR & SPI_FLAG_TXE) {
					SD_SPI->DR = 0xFF;
				}
				if(SD_SPI->SR & SPI_FLAG_RXNE) {
					response--;
					((uint8_t*)&data_left)[response] = (uint8_t)(SD_SPI->DR);
				}
			}while(response);

			if(crc == data_left) {
				SD_SPI_CE_DISABLE(SD);
				return(MSMD_AccessResult_Success);
			}
		}
	}
	SD_SPI_CE_DISABLE(SD);
	return(MSMD_AccessResult_DataError);
}




static MSMD_AccessResult SD_SPI_ReadBlock(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	SD_SPI_MSD *SD;
	if(MSD) {
		SD = (SD_SPI_MSD*)MSD;
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Busy;
		do {
			if(SD_SPI_ReadSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
				if(SD_SPI_ReadSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
					if(SD_SPI_ReadSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
						SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
						return(MSMD_AccessResult_DataError);
					}
				}
			}
			count--;
			data += MSD->read_sector_size;
			if(addr_L > ((uint32_t)(addr_L + MSD->read_sector_size))) {
				addr_H++;
			}
			addr_L += MSD->read_sector_size;
		}while(count);
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
		return(MSMD_AccessResult_Success);
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_WRITE_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_WriteSector(SD_SPI_MSD *SD, uint8_t *data, uint32_t addr_L, uint32_t addr_H) {
	uint8_t response = 0xFF;
	uint16_t crc = 0;
	uint32_t data_left = SD->SD_Desc.read_sector_size;
	uint8_t R2_second_byte;
	SPI_TypeDef *SD_SPI = SD->SD_SPI;
	
	SD_SPI_CE_ENABLE(SD);
	SD_SPI_SendCommand(SD_SPI, 24, addr_L);
	SD_SPI_Get1ByteToken(SD_SPI, &response);
	// SD card is ready and there was no error in the command
	if(!response) {
		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = 0xFF;

		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = 0xFE;

		while(data_left){
			while(!(SD_SPI->SR & SPI_FLAG_TXE));
			SD_SPI->DR = *data;
			crc = crc16(crc, *data);
			data_left--;
			data++;
		}

		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = ((uint8_t*)&crc)[1];
		
		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = ((uint8_t*)&crc)[0];

		SD_SPI_Get1ByteToken(SD_SPI, &response);
		if((response & 0x1F) == 0x5) {
			// wait till card sends busy signal
			do {
				if(SD_SPI->SR & SPI_FLAG_TXE) {
					SD_SPI->DR = 0xFF;
				}
				if(SD_SPI->SR & SPI_FLAG_RXNE) {
					response = (uint8_t)(SD_SPI->DR);
				}
			}while(!response);

			SD_SPI_CE_DISABLE(SD);

			SD_SPI_CE_ENABLE(SD);

			SD_SPI_SendCommand(SD_SPI, 13, 0);
			SD_SPI_Get1ByteToken(SD_SPI, &response);
			SD_SPI_Get1ByteToken(SD_SPI, &R2_second_byte);
			if(!response && !R2_second_byte) {
				SD_SPI_CE_DISABLE(SD);

				return(MSMD_AccessResult_Success);
			}
		}
	}
	SD_SPI_CE_DISABLE(SD);
	return(MSMD_AccessResult_DataError);
}
#endif




#if(MSM_WRITE_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_WriteBlock(MSMD_Desc *MSD, uint8_t *data, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	SD_SPI_MSD *SD;
	if(MSD) {
		SD = (SD_SPI_MSD*)MSD;
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Busy;
		do {
			if(SD_SPI_WriteSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
				if(SD_SPI_WriteSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
					if(SD_SPI_WriteSector(SD, data, addr_L, addr_H) == MSMD_AccessResult_DataError) {
						SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
						return(MSMD_AccessResult_DataError);
					}
				}
			}
			count--;
			data += MSD->write_sector_size;
			if(addr_L > ((uint32_t)(addr_L + MSD->write_sector_size))) {
				addr_H++;
			}
			addr_L += MSD->write_sector_size;
		}while(count);
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
		return(MSMD_AccessResult_Success);
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_VERIFY_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_VerifyBlock(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	if(MSD) {
		return(MSMD_AccessResult_Success);
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif



#if(MSM_ERASE_SUPPORT == 1)
static MSMD_AccessResult SD_SPI_EraseBlock(MSMD_Desc *MSD, uint32_t addr_L, uint32_t addr_H, uint32_t count) {
	SD_SPI_MSD *SD;
	SPI_TypeDef *SD_SPI;
	uint8_t response;
	if(MSD) {
		SD = (SD_SPI_MSD*)MSD;
		SD_SPI = SD->SD_SPI;
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Busy;
		response = 0xFF;
		SD_SPI_CE_ENABLE(SD);

		SD_SPI_SendCommand(SD_SPI, 32, addr_L);
		SD_SPI_Get1ByteToken(SD_SPI, &response);

		SD_SPI_CE_DISABLE(SD);
		// SD card is ready and there was no error in the command
		if(!response) {
			SD_SPI_CE_ENABLE(SD);

			SD_SPI_SendCommand(SD_SPI, 33, addr_L + (count * SD->SD_Desc.erase_sector_size));
			SD_SPI_Get1ByteToken(SD_SPI, &response);

			// SD card is ready and there was no error in the command
			if(!response) {
				// wait till card sends busy signal
				do {
					if(SD_SPI->SR & SPI_FLAG_TXE) {
						SD_SPI->DR = 0xFF;
					}
					if(SD_SPI->SR & SPI_FLAG_RXNE) {
						response = (uint8_t)(SD_SPI->DR);
					}
				}while(!response);

				SD_SPI_CE_DISABLE(SD);

				SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
				return(MSMD_AccessResult_Success);
			}
			SD_SPI_CE_DISABLE(SD);
		}
		SD->SD_Parameters.DeviceState = MSMD_AccessResult_Idle;
		return(MSMD_AccessResult_DataError);
	}
	return(MSMD_AccessResult_AccessForbidden);
}
#endif





static void SD_SPI_SendCommand(SPI_TypeDef *SD_SPI, uint8_t cmd_index, uint32_t argument) {
	uint8_t i = 0;
	uint8_t command[6];
	command[0] = 0x40 | (cmd_index & 0x3F);
	command[1] = (uint8_t)((argument >> 24) & 0xFF);
	command[2] = (uint8_t)((argument >> 16) & 0xFF);
	command[3] = (uint8_t)((argument >> 8) & 0xFF);
	command[4] = (uint8_t)(argument & 0xFF);
	i = crc7(i, command[0]);
	i = crc7(i, command[1]);
	i = crc7(i, command[2]);
	i = crc7(i, command[3]);
	i = crc7(i, command[4]);
	command[5] = i << 1;

	for(i=0; i < 6; i++) {
		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = command[i];
	}
}




static void SD_SPI_Get1ByteToken(SPI_TypeDef *SD_SPI, uint8_t *response) {
	uint32_t counter = 0xFFF;
	uint8_t resp = 0xFF;

	// wait for byte that is no FF
	// this indicate, that the card sent
	// first byte of response
	do {
		if(SD_SPI->SR & SPI_FLAG_TXE) {
			SD_SPI->DR = 0xFF;
		}
		if(SD_SPI->SR & SPI_FLAG_RXNE) {
			resp = (uint8_t)(SD_SPI->DR);
			counter--;
		}
	} while((resp == 0xFF) && counter);

	// save response and exit
	*response = resp;
}


static void SD_SPI_GetR3R7(SPI_TypeDef *SD_SPI, uint8_t *response) {
	uint32_t counter = 0xFFF;

	// wait for byte that is no FF
	// this indicate, that card sent
	// first byte of response
	SD_SPI_Get1ByteToken(SD_SPI, &response[4]);

	// receive next bytes of response
	if(response[4] != 0xFF) {
		counter = 4;
		do {
			if(SD_SPI->SR & SPI_FLAG_TXE) {
				SD_SPI->DR = 0xFF;
			}
			if(SD_SPI->SR & SPI_FLAG_RXNE) {
				counter--;
				response[counter] = (uint8_t)(SD_SPI->DR);
			}
		}while(counter);
	}
}



static int SD_SPI_CID_CSD_Get_data(SPI_TypeDef *SD_SPI, uint8_t *data) {
	uint8_t response;
	uint32_t counter = 15;
	uint8_t crc_7_calculated = 0;
	uint8_t crc_7_received = 0;
	// data
	do {
		if(SD_SPI->SR & SPI_FLAG_TXE) {
			SD_SPI->DR = 0xFF;
		}
		if(SD_SPI->SR & SPI_FLAG_RXNE) {
			counter--;
			response = (uint8_t)(SD_SPI->DR);
			data[counter] = response;
			crc_7_calculated = crc7(crc_7_calculated, response);
			
		}
	}while(counter);

	// crc7
	counter = 1;
	do {
		if(SD_SPI->SR & SPI_FLAG_TXE) {
			SD_SPI->DR = 0xFF;
		}
		if(SD_SPI->SR & SPI_FLAG_RXNE) {
			counter--;
			crc_7_received = (uint8_t)(SD_SPI->DR);
		}
	}while(counter);
	
	crc_7_received >>= 1;
	if(crc_7_received == crc_7_calculated) {
		return(0);
	}
	return(-1);
}



static int SD_SPI_GetCID(SD_SPI_MSD *SD, SD_CID_register *CID) {
	uint8_t response = 0xFF;
	uint32_t counter = 15;
	int result;
	uint8_t crc16_ignored;
	SPI_TypeDef *SD_SPI = SD->SD_SPI;
	
	SD_SPI_CE_ENABLE(SD);

	SD_SPI_SendCommand(SD_SPI, 10, 0);
	SD_SPI_Get1ByteToken(SD_SPI, &response);
	// SD card is ready and there was no error in the command
	if(!(response & 0x7F)) {
		SD_SPI_Get1ByteToken(SD_SPI, &response);
		// Single block read
		if(response == 0xFE) {
			result = SD_SPI_CID_CSD_Get_data(SD_SPI, (uint8_t*)CID);

			CID->MDT &= 0xFFF;

			counter = 2;
			do {
				if(SD_SPI->SR & SPI_FLAG_TXE) {
					SD_SPI->DR = 0xFF;
				}
				if(SD_SPI->SR & SPI_FLAG_RXNE) {
					counter--;
					crc16_ignored = (uint8_t)(SD_SPI->DR);
				}
			}while(counter);

			if(!result) {
				SD_SPI_CE_DISABLE(SD);
				return(0);

			}
		}
	}

	SD_SPI_CE_DISABLE(SD);
	return(-1);
}



static int SD_SPI_GetCSD(SD_SPI_MSD *SD, uint8_t *CSD) {
	uint8_t response = 0xFF;
	uint32_t counter = 15;
	int result;
	uint8_t crc16_ignored;
	SPI_TypeDef *SD_SPI = SD->SD_SPI;
	
	SD_SPI_CE_ENABLE(SD);

	SD_SPI_SendCommand(SD_SPI, 9, 0);
	SD_SPI_Get1ByteToken(SD_SPI, &response);
	// SD card is ready and there was no error in the command
	if(!response) {
		SD_SPI_Get1ByteToken(SD_SPI, &response);
		// Single block read
		if(response == 0xFE) {
			result = SD_SPI_CID_CSD_Get_data(SD_SPI, CSD);

			counter = 2;
			do {
				if(SD_SPI->SR & SPI_FLAG_TXE) {
					SD_SPI->DR = 0xFF;
				}
				if(SD_SPI->SR & SPI_FLAG_RXNE) {
					counter--;
					crc16_ignored = (uint8_t)(SD_SPI->DR);
				}
			}while(counter);

			if(!result) {
				SD_SPI_CE_DISABLE(SD);
				return(0);
			}
		}
	}

	SD_SPI_CE_DISABLE(SD);
	return(-1);
}




static uint8_t SD_SPI_Initialize(SD_SPI_MSD *SD) {
	uint8_t resp[5];
	int i;
	SPI_TypeDef *SD_SPI = SD->SD_SPI;

	SD_SPI_CE_DISABLE(SD);
	
	// set at least 74 clock cycles to prepare
	// all internal SD registers (we set a bit more)
	for(i=0; i < 11; i++) {
		while(!(SD_SPI->SR & SPI_FLAG_TXE));
		SD_SPI->DR = 0xFF;
	}
	
	// CE set to low
	SD_SPI_CE_ENABLE(SD);
	// send reset command (CMD0)
	SD_SPI_SendCommand(SD_SPI, 0, 0);
	SD_SPI_Get1ByteToken(SD_SPI, resp);
	SD_SPI_CE_DISABLE(SD);

	// small delay after reset
	for(i=0; i < 1000; i++) {
		asm("nop");
	}

	SD_SPI_CE_ENABLE(SD);
	SD_SPI_SendCommand(SD_SPI, 8, 0x1AA);
	SD_SPI_GetR3R7(SD_SPI, resp);
	SD_SPI_CE_DISABLE(SD);

	// valid response with correct pattern and correct supply voltage
	if((resp[0] == 0xAA) && (resp[1] == 1) && (resp[4] == 1)) {
		SD_SPI_CE_ENABLE(SD);
		SD_SPI_SendCommand(SD_SPI, 58, 0);
		SD_SPI_GetR3R7(SD_SPI, resp);
		SD_SPI_CE_DISABLE(SD);
		
		if((resp[4] != 0xFF) && (resp[2] || (resp[1] & 0x80))) {

			// CMD59
			SD_SPI_CE_ENABLE(SD);
			SD_SPI_SendCommand(SD_SPI, 59, 1);
			SD_SPI_GetR3R7(SD_SPI, resp);
			SD_SPI_CE_DISABLE(SD);

			do {
				SD_SPI_CE_ENABLE(SD);
				SD_SPI_SendCommand(SD_SPI, 55, 0);
				SD_SPI_Get1ByteToken(SD_SPI, resp);
				SD_SPI_CE_DISABLE(SD);
				SD_SPI_CE_ENABLE(SD);
				SD_SPI_SendCommand(SD_SPI, 41, 0x40000000);
				SD_SPI_Get1ByteToken(SD_SPI, resp);
				SD_SPI_CE_DISABLE(SD);//*/
				// Send CMD1
//				SD_SPI_CE_ENABLE();
//				SD_SPI_SendCommand(1, 0x40000000);
//				SD_SPI_Get1ByteToken(resp);
//				SD_SPI_CE_DISABLE();
			}while(resp[0] & 0x7f);
			
			// CMD58
			SD_SPI_CE_ENABLE(SD);
			SD_SPI_SendCommand(SD_SPI, 58, 0);
			SD_SPI_GetR3R7(SD_SPI, resp);
			SD_SPI_CE_DISABLE(SD);
			
			
		// TODO: CCS check
			
			return(0);
		}
	}
	// illegal command
	else if(resp[4] == 0xFF) {
		SD_SPI_CE_ENABLE(SD);
		SD_SPI_SendCommand(SD_SPI, 59, 1);
		SD_SPI_GetR3R7(SD_SPI, resp);
		SD_SPI_CE_DISABLE(SD);

		SD_SPI_CE_ENABLE(SD);
		SD_SPI_SendCommand(SD_SPI, 58, 0);
		SD_SPI_GetR3R7(SD_SPI, resp);
		SD_SPI_CE_DISABLE(SD);
		
		if((resp[4] != 0xFF) && (resp[2] || (resp[1] & 0x80))) {
			do {
				// Send CMD1
				SD_SPI_CE_ENABLE(SD);
				SD_SPI_SendCommand(SD_SPI, 1, 0);
				SD_SPI_Get1ByteToken(SD_SPI, resp);
				SD_SPI_CE_DISABLE(SD);
			}while(resp[0] & 0x1);
			return(0);
		}
	}
	return(1);
}
